"""Tools to help with the common problems of Forta bot development."""

import forta_toolkit.alerts
import forta_toolkit.indexing
import forta_toolkit.logging
import forta_toolkit.parsing
import forta_toolkit.profiling
import forta_toolkit.scraping



"""Profile the performances of a bot."""

import cProfile
import functools
import logging
import pstats
import time

# TIMEIT ######################################################################

def timeit(func: callable) -> callable:
    """Decorate a function to automatically log its execution time."""
    @functools.wraps(func)
    def __wrapper(*args, **kwargs):
        __start = time.perf_counter()
        __result = func(*args, **kwargs)
        __delta = 1000. * (time.perf_counter() - __start)
        logging.debug(f'{func.__name__} took {__delta:.9f} ms')
        return __result
    return __wrapper

# STATS #######################################################################

def profile(func: callable) -> callable:
    """Decorate a function to automatically collect performance statistics on its execution."""
    @functools.wraps(func)
    def __wrapper(*args, **kwargs):
        __profiler = cProfile.Profile()
        __profiler.enable()
        __result = func(*args, **kwargs)
        __profiler.disable()
        __profiler.dump_stats(f'{func.__name__}')
        return __result
    return __wrapper

# TESTS #######################################################################

@profile
def test_performances(func: callable, data: tuple) -> None:
    """Profile a function with the provided data."""
    (func(__d) for __d in data)

def display_performances(logpath: str) -> str:
    """Breakdown the performances of the full chain of execution in a table."""
    _p = pstats.Stats(logpath)
    return _p.strip_dirs().sort_stats('cumulative').print_stats()



"""Format blockchain logs."""

import forta_agent.transaction

import forta_toolkit.parsing.address

# TRANSACTION LOGS ############################################################

def parse_transaction_data(transaction: forta_agent.transaction.Transaction) -> dict:
    """Flatten and format all the data in a transaction log."""
    return {
        'hash': getattr(transaction, 'hash', '0x'),
        'from': forta_toolkit.parsing.address.format_with_checksum(getattr(transaction, 'from_', '')),
        'to': forta_toolkit.parsing.address.format_with_checksum(getattr(transaction, 'to', '')),
        'value': getattr(transaction, 'value', '0x'),
        'data': getattr(transaction, 'data', '0x'),}



"""Format transaction traces."""

import hexbytes

import forta_toolkit.parsing.address

# TRACES ######################################################################

def parse_trace_data(trace: dict, get_code: callable) -> dict:
    """Flatten and format all the data in a transaction trace."""
    # common
    __block = getattr(trace, 'block_number', getattr(trace, 'blockNumber', -1))
    __hash = getattr(trace, 'transaction_hash', getattr(trace, 'transactionHash', hexbytes.HexBytes('0x'))).hex()
    __type = getattr(trace, 'type', '')
    __action = getattr(trace, 'action', {})
    __result = getattr(trace, 'result', {})
    __value = getattr(__action, 'value', 0)
    __gas = getattr(__result, 'gas_used', getattr(__result, 'gasUsed', 0))
    # output
    __data = {
        'block': __block,
        'hash': __hash,
        'type': __type,
        'value': __value,
        'gas': __gas,
        'from': '',
        'to': '',
        'input': '',
        'output': ''}
    # call
    if __type == 'call':
        __data['type'] = getattr(__action, 'call_type', getattr(__action, 'callType', 'call')) # actually get the exact variant of the call
        __data['from'] = forta_toolkit.parsing.address.format_with_checksum(getattr(__action, 'from_', getattr(__action, 'from', '')))
        __data['to'] = forta_toolkit.parsing.address.format_with_checksum(getattr(__action, 'to', ''))
        __data['input'] = getattr(__action, 'input', hexbytes.HexBytes('0x')).hex()
        __data['output'] = getattr(__result, 'output', hexbytes.HexBytes('0x')).hex()
    # create
    if __type == 'create':
        __data['from'] = forta_toolkit.parsing.address.format_with_checksum(getattr(__action, 'from_', getattr(__action, 'from', '')))
        __data['to'] = forta_toolkit.parsing.address.format_with_checksum(getattr(__result, 'address', ''))
        __data['input'] = getattr(__action, 'init', hexbytes.HexBytes('0x')).hex()
        __data['output'] = getattr(__result, 'code', hexbytes.HexBytes('0x')).hex()
    # suicide
    if __type == 'suicide':
        __data['from'] = forta_toolkit.parsing.address.format_with_checksum(getattr(__action, 'address', ''))
        __data['to'] = forta_toolkit.parsing.address.format_with_checksum(getattr(__action, 'refund_address', getattr(__action, 'refundAddress', '')))
        __data['input'] = getattr(__action, 'balance', hexbytes.HexBytes('0x')).hex()
        __data['output'] = '0x'
    return __data
